@@ -19,6 +19,7 @@ use futures_util::StreamExt;
 use indicatif::{MultiProgress, ProgressBar, ProgressStyle};
 use log::{debug, trace};
 use reqwest::{header, Client, Response, StatusCode};
+use tokio::task::JoinSet;
 use xdg::BaseDirectories;
 use xz2::write::XzDecoder;
 
@@ -30,18 +31,16 @@ pub(crate) fn default_progress_template() -> &'static str {
     "{msg}: {spinner:.green} [{wide_bar:.cyan/blue}] {pos}/{len} ({per_sec}, {eta})"
 }
 
-#[derive(Debug, Clone, Copy)]
 pub(crate) enum CacheEntries {
     Excuses,
     Packages(Suite),
     Sources(Suite),
     FTBFSBugs(Codename),
     AutoRemovals,
-    Contents(Suite),
     Release(Suite),
 }
 
-#[derive(Debug, Clone, Copy, Eq, PartialEq)]
+#[derive(Clone, Copy, Eq, PartialEq)]
 pub(crate) enum CacheState {
     NoUpdate,
     FreshFiles,
@@ -54,7 +53,24 @@ enum Compressor {
     None,
 }
 
-#[derive(Debug, Clone)]
+#[derive(Debug)]
+struct DownloadInfo {
+    url: Cow<'static, str>,
+    destination: Cow<'static, str>,
+    compressor: Compressor,
+}
+
+impl DownloadInfo {
+    fn new(url: Cow<'static, str>, destination: Cow<'static, str>) -> Self {
+        Self {
+            url,
+            destination,
+            compressor: Compressor::None,
+        }
+    }
+}
+
+#[derive(Clone)]
 struct Downloader {
     always_download: bool,
     client: Client,
@@ -171,7 +187,9 @@ impl Downloader {
         } else if compressor == Compressor::Gz {
             let mut writer = GzDecoder::new(file);
             self.download_internal(res, &pb, &mut writer).await?;
-            writer.try_finish()?;
+            writer
+                .try_finish()
+                .with_context(|| format!("Failed to decompress {}", url))?;
         } else {
             self.download_internal(res, &pb, &mut file).await?;
         }
@@ -187,46 +205,28 @@ impl Downloader {
     }
 }
 
-fn excuses_urls() -> Vec<(Cow<'static, str>, Compressor, Cow<'static, str>)> {
-    vec![(
-        "https://release.debian.org/britney/excuses.yaml.gz".into(),
-        Compressor::Gz,
-        "excuses.yaml".into(),
-    )]
+fn excuses_urls() -> Vec<DownloadInfo> {
+    vec![DownloadInfo {
+        url: "https://release.debian.org/britney/excuses.yaml.gz".into(),
+        compressor: Compressor::Gz,
+        destination: "excuses.yaml".into(),
+    }]
 }
 
-fn ftbfs_bugs_urls(codename: Codename) -> Vec<(Cow<'static, str>, Compressor, Cow<'static, str>)> {
-    vec![(
+fn ftbfs_bugs_urls(codename: Codename) -> Vec<DownloadInfo> {
+    vec![DownloadInfo::new (
         format!("https://udd.debian.org/bugs/?release={}&ftbfs=only&merged=ign&done=ign&rc=1&sortby=id&sorto=asc&format=yaml", codename).into(),
-        Compressor::None,
         format!("udd-ftbfs-bugs-{}.yaml", codename).into()
     )]
 }
 
-fn auto_removals_urls() -> Vec<(Cow<'static, str>, Compressor, Cow<'static, str>)> {
-    vec![(
+fn auto_removals_urls() -> Vec<DownloadInfo> {
+    vec![DownloadInfo::new(
         "https://udd.debian.org/cgi-bin/autoremovals.yaml.cgi".into(),
-        Compressor::None,
         "autoremovals.yaml".into(),
     )]
 }
 
-#[derive(Debug)]
-pub(crate) struct Cache {
-    base_directory: BaseDirectories,
-    downloader: Downloader,
-    archive_mirror: String,
-    unstable: release::Release,
-    testing: release::Release,
-    stable: release::Release,
-    oldstable: release::Release,
-    experimental: release::Release,
-    stable_proposed_updates: release::Release,
-    oldstable_proposed_updates: release::Release,
-    stable_backports: release::Release,
-    // oldstable_backports: release::Release,
-}
-
 fn empty_release() -> release::Release {
     release::Release {
         origin: Default::default(),
@@ -244,6 +244,21 @@ fn empty_release() -> release::Release {
     }
 }
 
+pub(crate) struct Cache {
+    base_directory: BaseDirectories,
+    downloader: Downloader,
+    archive_mirror: String,
+    unstable: release::Release,
+    testing: release::Release,
+    stable: release::Release,
+    oldstable: release::Release,
+    experimental: release::Release,
+    stable_proposed_updates: release::Release,
+    oldstable_proposed_updates: release::Release,
+    stable_backports: release::Release,
+    // oldstable_backports: release::Release,
+}
+
 impl Cache {
     pub async fn new(force_download: bool, archive_mirror: &str) -> Result<Self> {
         let mut cache = Self {
@@ -336,63 +351,35 @@ impl Cache {
         )
     }
 
-    fn contents_urls(
-        &self,
-        suite: Suite,
-    ) -> Vec<(Cow<'static, str>, Compressor, Cow<'static, str>)> {
+    fn packages_urls(&self, suite: Suite) -> Vec<DownloadInfo> {
         self.architectures_for_suite(suite)
             .into_iter()
-            .map(|architecture| {
-                (
-                    self.lookup_url(suite, &format!("main/Contents-{}.gz", architecture))
-                        .into(),
-                    Compressor::Gz,
-                    format!("Contents_{}_{}", suite, architecture).into(),
-                )
+            .map(|architecture| DownloadInfo {
+                url: self
+                    .lookup_url(suite, &format!("main/binary-{}/Packages.xz", architecture))
+                    .into(),
+                compressor: Compressor::Xz,
+                destination: format!("Packages_{}_{}", suite, architecture).into(),
             })
             .collect()
     }
 
-    fn packages_urls(
-        &self,
-        suite: Suite,
-    ) -> Vec<(Cow<'static, str>, Compressor, Cow<'static, str>)> {
-        self.architectures_for_suite(suite)
-            .into_iter()
-            .map(|architecture| {
-                (
-                    self.lookup_url(suite, &format!("main/binary-{}/Packages.xz", architecture))
-                        .into(),
-                    Compressor::Xz,
-                    format!("Packages_{}_{}", suite, architecture).into(),
-                )
-            })
-            .collect()
+    fn source_urls(&self, suite: Suite) -> Vec<DownloadInfo> {
+        vec![DownloadInfo {
+            url: self.lookup_url(suite, "main/source/Sources.xz").into(),
+            compressor: Compressor::Xz,
+            destination: format!("Sources_{}", suite).into(),
+        }]
     }
 
-    fn source_urls(&self, suite: Suite) -> Vec<(Cow<'static, str>, Compressor, Cow<'static, str>)> {
-        vec![(
-            self.lookup_url(suite, "main/source/Sources.xz").into(),
-            Compressor::Xz,
-            format!("Sources_{}", suite).into(),
-        )]
-    }
-
-    fn release_urls(
-        &self,
-        suite: Suite,
-    ) -> Vec<(Cow<'static, str>, Compressor, Cow<'static, str>)> {
-        vec![(
+    fn release_urls(&self, suite: Suite) -> Vec<DownloadInfo> {
+        vec![DownloadInfo::new(
             format!("{}/dists/{}/Release", self.archive_mirror, suite).into(),
-            Compressor::None,
             format!("Release_{}", suite).into(),
         )]
     }
 
-    fn cache_entries_to_urls_dests(
-        &self,
-        entries: &[CacheEntries],
-    ) -> Result<Vec<(Cow<'static, str>, Compressor, PathBuf)>> {
+    fn cache_entries_to_urls_dests(&self, entries: &[CacheEntries]) -> Vec<DownloadInfo> {
         entries
             .iter()
             .flat_map(|entry| {
@@ -402,19 +389,15 @@ impl Cache {
                     CacheEntries::Sources(suite) => self.source_urls(*suite),
                     CacheEntries::FTBFSBugs(codename) => ftbfs_bugs_urls(*codename),
                     CacheEntries::AutoRemovals => auto_removals_urls(),
-                    CacheEntries::Contents(suite) => self.contents_urls(*suite),
                     CacheEntries::Release(suite) => self.release_urls(*suite),
                 }
                 .into_iter()
             })
-            .map(|(url, compressor, dest)| {
-                Ok((url, compressor, self.get_cache_path(dest.as_ref())?))
-            })
             .collect()
     }
 
     pub async fn download(&self, entries: &[CacheEntries]) -> Result<CacheState> {
-        let urls_and_dests = self.cache_entries_to_urls_dests(entries)?;
+        let urls_and_dests = self.cache_entries_to_urls_dests(entries);
         trace!(
             "Scheduling {} URLs to download: {:?}",
             urls_and_dests.len(),
@@ -422,31 +405,28 @@ impl Cache {
         );
 
         let mp = MultiProgress::new();
-        let join_handles: Vec<_> = urls_and_dests
-            .into_iter()
-            .map(|(url, compressor, dest)| {
-                let downloader = self.downloader.clone();
-                let mp = mp.clone();
-                tokio::spawn(async move {
-                    debug!("Starting task to download {}", url);
-                    downloader.download_file(&url, dest, compressor, mp).await
-                })
-            })
-            .collect();
+        let mut join_handles = JoinSet::new();
+        for download_info in urls_and_dests {
+            let dest = self.get_cache_path(download_info.destination.as_ref())?;
+            let downloader = self.downloader.clone();
+            let mp = mp.clone();
+            join_handles.spawn(async move {
+                debug!("Starting task to download {}", download_info.url);
+                downloader
+                    .download_file(&download_info.url, dest, download_info.compressor, mp)
+                    .await
+            });
+        }
 
         let mut state = Ok(CacheState::NoUpdate);
-        for handle in join_handles {
-            match handle.await {
+        while let Some(res) = join_handles.join_next().await {
+            match res {
                 Ok(download_result) => match download_result {
-                    Ok(CacheState::FreshFiles) => {
-                        if state.is_ok() {
-                            state = Ok(CacheState::FreshFiles);
-                        }
-                    }
+                    new_state @ Ok(CacheState::FreshFiles) => state = state.and(new_state),
                     Err(err) => state = Err(err),
                     _ => {}
                 },
-                Err(err) => state = Err(err.into()),
+                Err(err) => state = Err(err).context("Failed to join task"),
             };
         }
         state
@@ -484,6 +464,10 @@ impl Cache {
         )?))
     }
 
+    pub fn get_package_path(&self, suite: Suite, architecture: Architecture) -> Result<PathBuf> {
+        self.get_cache_path(format!("Packages_{}_{}", suite, architecture))
+    }
+
     pub fn get_package_paths(&self, suite: Suite, with_all: bool) -> Result<Vec<PathBuf>> {
         let mut all_paths = vec![];
         for architecture in self.architectures_for_suite(suite) {
@@ -491,7 +475,7 @@ impl Cache {
                 continue;
             }
 
-            all_paths.push(self.get_cache_path(format!("Packages_{}_{}", suite, architecture))?);
+            all_paths.push(self.get_package_path(suite, architecture)?);
         }
         Ok(all_paths)
     }
@@ -500,17 +484,6 @@ impl Cache {
         self.get_cache_path(format!("Sources_{}", suite))
     }
 
-    pub fn get_content_paths(&self, suite: Suite) -> Result<Vec<(Architecture, PathBuf)>> {
-        let mut all_paths = vec![];
-        for architecture in self.architectures_for_suite(suite) {
-            all_paths.push((
-                architecture,
-                self.get_cache_path(format!("Contents_{}_{}", suite, architecture))?,
-            ));
-        }
-        Ok(all_paths)
-    }
-
     // Architectures for a suite (including Arch: all)
     pub fn architectures_for_suite(&self, suite: Suite) -> Vec<Architecture> {
         match suite {
@@ -531,6 +504,8 @@ pub fn source_skip_binnmu(source: &str) -> bool {
                 || source.starts_with("linux-")
                 || source.starts_with("shim-")
                 || source.starts_with("fwupd-")))
+        || (source.contains("cross")
+            && (source.starts_with("gcc-") || source.starts_with("binutils-")))
 }
 
 #[cfg(test)]
