@@ -1,3 +1,6 @@
+// Copyright 2022-2023 Sebastian Ramacher
+// SPDX-License-Identifier: GPL-3.0-or-later
+
 use std::{
     collections::HashMap,
     fmt::{Display, Formatter},
@@ -53,7 +56,7 @@ pub struct UDDBug {
     pub title: String,
 }
 
-#[derive(Debug, Default)]
+#[derive(Default)]
 pub struct UDDBugs {
     bugs: Vec<UDDBug>,
     source_index: HashMap<String, Vec<usize>>,
